/**
 ** A Virtual Host Monitor
 **
 ** Copyright 2009 Gary Wuertz gary@issiweb.com
 **
 ** This program is free software: you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
 ** the Free Software Foundation, either version 3 of the License, or
 ** (at your option) any later version.
 **
 ** This program is distributed in the hope that it will be useful,
 ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ** GNU General Public License for more details.
 **
 ** You should have received a copy of the GNU General Public License
 ** along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <algorithm>
#include <stdexcept>
#include <iostream>

#include <errno.h>
#include <glob.h>
#include <sys/inotify.h>
#include <string.h>

#include "monitor.h"

#define EVENT_SIZE  (sizeof (struct inotify_event))
#define BUF_LEN     (1024 * (EVENT_SIZE + 16))
#define PATH_LEN    256

using namespace std;
/**
 * Sort functions for the display
 */
bool byAccess(const vhost& a, const vhost& b)
{
  if (a.ldtm == b.ldtm) return a.hits > b.hits;
  return a.ldtm > b.ldtm;
}
bool byHits(const vhost& a, const vhost& b)
{
  if (a.hits == b.hits) return a.ldtm > b.ldtm;
  return a.hits > b.hits;
}
bool byName(const vhost& a, const vhost& b)
{
  string na = a.fields[a.nidx];
  string nb = b.fields[b.nidx];
  if (na == nb) return a.ldtm > b.ldtm;
  return na < nb;
}
/**
 * Return the hosts list
 */
vhost  * Monitor::getHosts(topvhost *app)
{
  int i, hits;
  long meter=0;

  for(i=hits=0;i<this->nhosts;i++) {
    if (this->hosts[i].dirty && !this->hosts[i].setClean(app))
      this->throwError("clean:");
    this->hlist[i] = this->hosts[i];
    hits  += this->hosts[i].hits;
    meter += this->hosts[i].meter;
    }
  switch (app->order) {
    case topvhost::BY_ACCESS:
      std::sort(this->hlist,this->hlist+this->nhosts,byAccess);
      break;
    case topvhost::BY_HITS:
      std::sort(this->hlist,this->hlist+this->nhosts,byHits);
      break;
    case topvhost::BY_NAME:
      std::sort(this->hlist,this->hlist+this->nhosts,byName);
      break;
    }
  this->hits    = hits;
  this->meter   = meter;
  app->total    = this->nhosts;
  return this->hlist;
}
/**
 * Get input from application or watchlist
 */
bool  Monitor::getInput(topvhost *app)
{
  switch(app->input()) {
    case 'q':
      return false;
    case 'r':
      for(int i=0;i<this->nhosts;i++)
        this->hosts[i].dirty = true;
      app->paint = false;
    }
  if (app->paint)
    app->paint = false;
  this->update(app);
  return true;
}
/**
 * Get the watch list - the list of domains and associated log files. The list
 * is constructed from:
 *
 *   1) "glob_src"   = a file system glob (with optional "glog_omit" exclusion)
 *   2) "file_src"   = the top level section of the specified config file
 *   3) "config_src" = the name of a section in the current config file
 *
 * For configuration files the syntax is <domain name>=<log file path>
 */
watchList* Monitor::getWatchList(Config * conf)
{
  watchList*  wl      = new watchList();
  const char* xfile   = conf->get("file_src").c_str();
  const char* pattern = conf->get("glob_src").c_str();
  const char* xsect   = conf->get("config_src").c_str();
  map<string,string>::iterator iter;
  int         len;

  if (strlen(pattern)>0) {            // handle glob as source
    char        path[PATH_LEN];
    glob_t      p;
    char**      w;
    const char* exclude = conf->get("glob_omit").c_str();
    int         i, offset, overhead, skip;
    string      sl;

    path[PATH_LEN-1] = 0;
    skip    = strlen(exclude);
    snprintf(path,PATH_LEN-1,pattern, "*");
    if (glob(path, GLOB_NOSORT, NULL, &p))
      this->throwError("glob:");
    offset = strstr(pattern,"%s") - pattern;
    overhead = strlen(pattern)-2;
    w = p.gl_pathv;
    for(i = 0;i<(int)p.gl_pathc;i++) {
      len = strlen(w[i]);
      if (skip>0 && len>skip && !strcmp(w[i]+len-skip,exclude)) continue;
      len -= overhead;
      strncpy(path,w[i]+offset,len);
      path[len] = 0;
      sl = string(path);
      snprintf(path,PATH_LEN-1,pattern,sl.c_str());
      wl->watch[sl] = path;
      }
    globfree(&p);
    }
  if (strlen(xfile)) {                  // Handle external file as source
    Config *xconf = new Config(xfile);
    map<string,string>* section = xconf->getSection("");
    for(iter=section->begin();iter!=section->end();++iter)
        wl->watch[iter->first] = iter->second;
    }
  if (strlen(xsect)>0) {              // Handle section within config as source
    map<string,string>* section = conf->getSection(xsect);
    for(iter=section->begin();iter!=section->end();++iter)
        wl->watch[iter->first] = iter->second;
    }
  wl->namelen = 0;
  for(iter=wl->watch.begin();iter!=wl->watch.end();++iter) {
    if ((len=iter->first.length())>wl->namelen)
      wl->namelen = len;
    }
  if (wl->namelen<1)
    throw runtime_error("Empty watch list");
  return wl;
}
/**
 * Initialize the monitor
 */
void  Monitor::init(topvhost *app)
{
  vhost * inst;
  int   i, nfiles;
  unsigned j,k;
  string s;
  watchList* wl;
  map<string,string>::iterator iter;

  wl = this->getWatchList(app->conf);
  this->nhosts  = nfiles = (int) wl->watch.size();
  this->notify = inotify_init();
  if (this->notify < 0)
    this->throwError("notify_init");
  this->hosts  = inst = new vhost[nfiles];
  this->hlist  = new vhost[nfiles];
  k = vhost::OUTPUT_FIELDS + app->needs.size();
  for(i=0,iter=wl->watch.begin();iter!=wl->watch.end();++iter,i++) {
    for(j=0;j<k;j++)
      inst[i].fields.push_back("");
    inst[i].fields[vhost::FILE_FIELD]   = iter->second;
    inst[i].fields[inst[i].nidx = app->nidx] = iter->first;
    inst[i].wd  = inotify_add_watch(this->notify,inst[i].fields[1].c_str(),IN_MODIFY);
    this->watch[inst[i].wd] = i;
    if (!inst[i].setDirty(app->options.init))
      this->throwError("stat:");
    }
  if (app->options.verify) {
    vhost *inst = this->getHosts(app);
    cout << "Logs Monitored" << endl;
    for(int i=0;i<nhosts;i++)
      inst[i].verify();
    }
  else if (!app->options.diagnose)
    app->begin(this->nhosts,wl->namelen);
}
/**
 * Format page header using a sprintf like specification using the place holders:
 *
 *  %H    = hour
 *  %I    = bytes read
 *  %M    = minute
 *  %S    = second
 *  %Y    = year
 *  %d    = day
 *  %m    = month
 *  %n    = total vhosts
 *  %o    = total hits
 *  %t,%T = elapsed time (hh:mm:ss or n)
 */
void Monitor::paintHeader(topvhost *app)
{
  time_t  seconds = time(NULL);
  struct  tm *ts  = localtime(&seconds);
  int i;unsigned j;

  if (app->passes++==0)
    app->start = seconds;
  app->wmove(0,0);
  for(j=0;j<app->hdr.size();j++) {
    string token = app->hdr[j].format;
    switch(app->hdr[j].field) {
      case ' ':
        app->wputs(token.c_str());
        break;
      case 'H':   app->wprint(token,ts->tm_hour);
        break;
      case 'I':
          if (token=="%#") {
            int factors[] = {1, ' ', 1024, 'K', 1024*1024, 'M', 1024*1024*1024, 'G'};
            for (i=0;i<=2 && this->meter>=static_cast<long>(factors[i+2]);i+=2) ;
            app->wprint("%6.3",static_cast<float>(this->meter)/factors[i]);
            app->wput(factors[i+1]);
            break;
            }
        app->wprint(token,this->meter);
        break;
      case 'M':   app->wprint(token,ts->tm_min);
        break;
      case 'S':   app->wprint(token,ts->tm_sec);
        break;
      case 'T':   app->wprint(token,static_cast<int>(seconds - app->start));
        break;
      case 'Y':   app->wprint(token,ts->tm_year);
        break;
      case 'd':   app->wprint(token,ts->tm_mday);
        break;
      case 'm':   app->wprint(token,ts->tm_mon+1);
        break;
      case 'n':   app->wprint(token,this->nhosts);
        break;
      case 'o':   app->wprint(token,this->hits);
        break;
      case 't':
        i = static_cast<int>(seconds - app->start);
        int hms[3] = {(i/3600)%60,(i/60)%60,i%60};
        app->wprint("%02",hms[0]);
        app->wput(':');
        app->wprint("%02",hms[1]);
        app->wput(':');
        app->wprint("%02",hms[2]);
        break;
      }
    }
}
/**
 * Terminate the monitor
 */
void Monitor::quit(topvhost *app)
{
  app->end();
  if (this->notify>0) {
    close(this->notify);
    this->notify = 0;
    }
}
/**
 * Terminate with error message
 */
void Monitor::throwError(const char* tag)
{
  string msg = tag;
  char buf[256];
  strerror_r(errno, buf, sizeof(buf)-1);
  msg += buf;
  throw runtime_error(msg.c_str());
}
/**
 * Update the list - update the last date time modified and set the dirty bit
 * for inotify events received
 */
void Monitor::update(topvhost *app)
{
  char buf[BUF_LEN];
  struct inotify_event *event;
  fd_set read_fd;
  struct timeval tv;
  int len, i, idx, rc;

  FD_ZERO(&read_fd);
  FD_SET(this->notify,&read_fd);
  tv.tv_sec = tv.tv_usec = 0;
  rc = select(this->notify+1, &read_fd, NULL, NULL, &tv);
  if (rc<=0) {
    if (rc<0 && errno != EINTR)
      this->throwError("select:");
    }
  else {
    len = read (this->notify, buf, BUF_LEN);
    if (len<0 && errno!=EINTR)
      this->throwError("read:");
    for(i=0;i<len;i+=EVENT_SIZE + event->len) {
      event = (struct inotify_event *) &buf[i];
      idx = this->watch[event->wd];
      if ((event->mask & IN_MODIFY) != 0 && !this->hosts[idx].setDirty())
        this->throwError("stat:");
      }
    }
}
